# Autogenerated by devscripts/update-version.py

__version__ = '2025.09.23'

RELEASE_GIT_HEAD = '2e81e298cdce23afadb06a95836284acb38f7018'

VARIANT = None

UPDATE_HINT = None

CHANNEL = 'stable'

ORIGIN = 'yt-dlp/yt-dlp'

_pkg_version = '2025.09.23'
